import {navigation} from '../../../../src/shared/utils/navigation.utils';
import * as utils from '../../../utils';

const url = 'https://earthquake.usgs.gov/ws/nshmp/data/gmm/spectra';

describe('Response Spectra', () => {
  beforeEach(() => {
    cy.viewport('macbook-16');
    cy.intercept(url).as('usage');
    cy.visit('/gmm/spectra');
    cy.wait('@usage');
  });

  describe('Check nshmp-template', () => {
    utils.hasNshmpTemplate(navigation());
  });

  describe('Application', () => {
    utils.hasControlPanel();
    utils.hasPlotContent();
    utils.toggleControlPanel();
    utils.togglePlotPanel();
    utils.toggleSettingPanel();
    utils.changePlotSettings('.spectra-plot', '.spectra-settings');
    utils.changePlotSettings('.sigma-plot', '.sigma-settings');
    utils.hasGmmMenu();

    it('Can set GMM menus', () => {
      setGmmMenus();
    });

    it('Can set event parameters', () => {
      setEventParameters();
    });

    it('Can set geometry parameters', () => {
      utils.setSourceGeometryParameters();
    });

    it('Can set site parameters', () => {
      utils.setSiteParameters();
    });

    it('Can set path parameters', () => {
      setPathParameters();
    });

    it('Can reset control panel inputs', () => {
      utils.canResetControlPanel(() => setControlPanel());
    });

    it('Calls service and plots response', () => {
      const buttons: utils.ControlPanelButtons = {
        hasResetButton: false,
        hasServiceCallButton: true,
        hasSubmitButton: true,
      };
      utils.controlPanelButtonsNotEnabled(buttons);

      setControlPanel();
      utils.submitFormCheckPlots({
        firstPlotClass: '.spectra-plot',
        secondPlotClass: '.sigma-plot',
      });

      utils.controlPanelButtonsEnabled(buttons);
    });

    describe('Application Tabs', () => {
      it('Has plot tab', () => {
        cy.get('plotly-plot');

        utils.hasApplicationTab('.medians-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasApplicationTab('.plots-tab');

        cy.get('plotly-plot');
      });

      it('Has medians tab', () => {
        cy.get('plotly-plot');

        utils.exportDataTableNotExist();
        utils.dataTableNotExist();

        setControlPanel();

        utils.submitFormCheckPlots({
          firstPlotClass: '.spectra-plot',
          secondPlotClass: '.sigma-plot',
        });

        utils.hasApplicationTab('.medians-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasExportDataTable('spectra-medians.csv');
        utils.hasDataTable();
      });

      it('Has sigmas tab', () => {
        cy.get('plotly-plot');

        utils.exportDataTableNotExist();
        utils.dataTableNotExist();

        setControlPanel();

        utils.submitFormCheckPlots({
          firstPlotClass: '.spectra-plot',
          secondPlotClass: '.sigma-plot',
        });

        utils.hasApplicationTab('.sigmas-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasExportDataTable('spectra-sigmas.csv');
        utils.hasDataTable();
      });
    });
  });
});

function setControlPanel() {
  setGmmMenus();
  setEventParameters();
  utils.setSourceGeometryParameters();
  setPathParameters();
  utils.setSiteParameters();
}

function setGmmMenus() {
  cy.get('nshmp-template-control-panel')
    .find('form')
    .should('be.visible')
    .as('form');

  cy.get('@form').find('.multi-select-parameter-menu').click();
  utils.selectMatOption('gmm');

  cy.get('@form')
    .find('gmm-lib-gmm-menu')
    .find('select')
    .select([0, 1, 2, 3, 4]);
}

function setPathParameters() {
  cy.get('nshmp-template-control-panel')
    .find('.rx-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('20.0')
    .should('have.value', '20.0');

  cy.get('nshmp-template-control-panel')
    .find('.rrup-input')
    .find('input')
    .should('be.visible')
    .should('not.be.enabled');

  cy.get('nshmp-template-control-panel')
    .find('.rjb-input')
    .find('input')
    .should('be.visible')
    .should('not.be.enabled');
}

function setEventParameters() {
  utils.setMagnitude();

  cy.get('nshmp-template-control-panel')
    .find('.rake-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('90')
    .should('have.value', '90');
}
