import {DisaggTarget} from '@ghsc/nshmp-lib-ng/hazard';

import {navigation} from '../../../../src/shared/utils/navigation.utils';
import * as utils from '../../../utils';

const url = 'https://earthquake.usgs.gov/ws/nshmp/*/dynamic/disagg';

describe('Disagg Application', () => {
  beforeEach(() => {
    cy.viewport('macbook-16');
    cy.intercept(url).as('usage');
    cy.visit('/hazard/disagg');
    cy.wait('@usage');
  });

  describe('Check nshmp-template', () => {
    utils.hasNshmpTemplate(navigation());
  });

  describe('Application', () => {
    utils.hasControlPanel();
    utils.hasPlotContent();
    utils.toggleControlPanel();
    utils.togglePlotPanel();
    utils.hasHazardModelMenu();

    it('Can set control panel', () => {
      setControlPanel();
    });

    it('Downloads data as csv', () => {
      setControlPanel();
      utils.submitFormCheckPlot({
        plotClass: 'hazard-lib-disagg-plot',
      });

      cy.get('nshmp-template-main-content')
        .find('.export-button')
        .scrollIntoView()
        .click();
      utils.hasDownloadedFile('disagg-Total.csv');
    });

    it('Downloads summary text', () => {
      setControlPanel();
      utils.submitFormCheckPlot({
        plotClass: 'hazard-lib-disagg-plot',
      });

      cy.get('nshmp-template-main-content')
        .find('.summary')
        .find('.summary-report')
        .scrollIntoView()
        .find('button')
        .click();
      utils.hasDownloadedFile('disagg-Total.csv');
    });

    it('Calls service and plots response', () => {
      const buttons: utils.ControlPanelButtons = {
        hasResetButton: false,
        hasServiceCallButton: true,
        hasSubmitButton: true,
      };
      utils.controlPanelButtonsNotEnabled(buttons);

      setControlPanel();

      cy.get('.cdk-overlay-container').as('overlay');

      cy.get('nshmp-template-main-content')
        .find('.component-select')
        .find('mat-select')
        .should('be.visible')
        .click();
      cy.get('@overlay').find('mat-option').should('not.exist');

      cy.get('nshmp-template-main-content')
        .find('.export-button')
        .should('be.visible')
        .should('not.be.enabled');

      cy.get('nshmp-template-main-content')
        .find('.summary')
        .find('app-disagg-summary')
        .scrollIntoView()
        .should('not.be.visible');

      cy.get('nshmp-template-main-content')
        .find('.summary')
        .find('app-disagg-contributors')
        .scrollIntoView()
        .should('not.be.visible');

      utils.submitFormCheckPlot({
        plotClass: 'hazard-lib-disagg-plot',
      });

      cy.get('nshmp-template-main-content')
        .find('.component-select')
        .scrollIntoView()
        .find('mat-select')
        .should('be.visible')
        .click();
      cy.get('@overlay').find('mat-option').should('have.length.above', 1);
      utils.selectMatOption('Total');

      cy.get('nshmp-template-main-content')
        .find('.export-button')
        .scrollIntoView()
        .should('be.visible')
        .should('be.enabled')
        .click();

      cy.get('nshmp-template-main-content')
        .find('.summary')
        .find('app-disagg-summary')
        .scrollIntoView()
        .should('be.visible');

      cy.get('nshmp-template-main-content')
        .find('.summary')
        .find('app-disagg-contributors')
        .scrollIntoView()
        .should('be.visible');

      utils.controlPanelButtonsEnabled(buttons);
    });

    describe('Disagg targets', () => {
      const {IML, RETURN_PERIOD} = DisaggTarget;

      const returnPeriodRadioButtonSelector =
        'mat-radio-button[ng-reflect-value=RETURN_PERIOD]';
      const returnPeriodSelector =
        '.disagg-target-return-period mat-form-field';
      const commonReturnPeriodSelector =
        '.disagg-target-return-period mat-select';
      const imlRadioButtonSelector = 'mat-radio-button[ng-reflect-value=IML]';
      const imlSelector = '.disagg-target-iml mat-form-field';
      const maxDirectionSelector = '.disagg-target-iml mat-checkbox';

      /**
       * Expect the controls for only the given disagg target to be active
       * @param expected The name of the disagg target
       */
      function checkDisaggTarget(expected: DisaggTarget) {
        // return period
        cy.get(returnPeriodRadioButtonSelector)
          .should(
            expected === RETURN_PERIOD ? 'have.class' : 'not.have.class',
            'mat-radio-checked',
          )
          .find('input')
          .should(expected === RETURN_PERIOD ? 'be.checked' : 'not.be.checked');
        cy.get(returnPeriodSelector)
          .should(
            expected === RETURN_PERIOD ? 'not.have.class' : 'have.class',
            'mat-form-field-disabled',
          )
          .find('input')
          .should(expected === RETURN_PERIOD ? 'be.enabled' : 'be.disabled');
        cy.get(commonReturnPeriodSelector).should(
          expected === RETURN_PERIOD ? 'not.have.class' : 'have.class',
          'mat-select-disabled',
        );

        // IML
        cy.get(imlRadioButtonSelector)
          .should(
            expected === IML ? 'have.class' : 'not.have.class',
            'mat-radio-checked',
          )
          .find('input')
          .should(expected === IML ? 'be.checked' : 'not.be.checked');
        cy.get(imlSelector)
          .should(
            expected === IML ? 'not.have.class' : 'have.class',
            'mat-form-field-disabled',
          )
          .find('input')
          .should(expected === IML ? 'be.enabled' : 'be.disabled');
        cy.get(maxDirectionSelector)
          .should(
            expected === IML ? 'not.have.class' : 'have.class',
            'mat-checkbox-disabled',
          )
          .find('input')
          .should(expected === IML ? 'be.enabled' : 'be.disabled');
      }

      describe('Return period', () => {
        // pretty much all of the existing tests work with return period
        // here we just make sure its disagg target-specific controls work

        it('is selected by default', () => {
          checkDisaggTarget(RETURN_PERIOD);
        });

        it('can be selected using the radio button', () => {
          cy.get(imlRadioButtonSelector).click();
          cy.get(returnPeriodRadioButtonSelector).click();

          checkDisaggTarget(RETURN_PERIOD);
        });

        it('unchecks the max direction checkbox when selected', () => {
          cy.get(imlRadioButtonSelector).click();
          cy.get(maxDirectionSelector).click();
          cy.get(returnPeriodRadioButtonSelector).click();

          cy.get(maxDirectionSelector)
            .should('not.have.class', 'mat-checkbox-checked')
            .find('input')
            .should('not.be.checked');
        });
      });

      describe('IML', () => {
        it('can be selected using the radio button', () => {
          cy.get(imlRadioButtonSelector).click();

          checkDisaggTarget(IML);
        });

        it('can type a number into the field', () => {
          cy.get(imlRadioButtonSelector).click();
          cy.get(imlSelector)
            .type('3.14')
            .find('input')
            .should('have.value', '3.14');
        });

        it('can check the max direction checkbox', () => {
          cy.get(imlRadioButtonSelector).click();
          cy.get(maxDirectionSelector)
            .click()
            .should('have.class', 'mat-checkbox-checked')
            .find('input')
            .should('be.checked');
        });

        it('disables and unchecks the max direction checkbox when PGV is selected', () => {
          cy.get(imlRadioButtonSelector).click();
          cy.get(maxDirectionSelector).click();
          cy.get('.imt-select').click();
          utils.selectMatOption('PGV');
          cy.get(maxDirectionSelector)
            .should('have.class', 'mat-checkbox-disabled')
            .should('not.have.class', 'mat-checkbox-checked')
            .find('input')
            .should('be.disabled')
            .should('not.be.checked');
        });

        it('filling out the form then switching to IML invalidates the form', () => {
          const buttons: utils.ControlPanelButtons = {
            hasResetButton: false,
            hasServiceCallButton: true,
            hasSubmitButton: true,
          };

          setControlPanel();
          cy.get(imlRadioButtonSelector).click();
          utils.controlPanelButtonsNotEnabled(buttons);
        });

        it('can call the service after filling in IML', () => {
          const buttons: utils.ControlPanelButtons = {
            hasResetButton: false,
            hasServiceCallButton: true,
            hasSubmitButton: true,
          };

          setControlPanel();
          cy.get(imlRadioButtonSelector).click();
          cy.get(imlSelector).type('3.14');
          utils.submitFormCheckPlot({
            plotClass: 'hazard-lib-disagg-plot',
          });
          utils.controlPanelButtonsEnabled(buttons);
        });

        it('calls the service with a modified IML when max-direction scaling is enabled', () => {
          setControlPanel();
          cy.get('.imt-select').click();
          utils.selectMatOption('SA0P02');
          cy.get(imlRadioButtonSelector).click();
          cy.get(imlSelector).type('6');
          cy.get(maxDirectionSelector).click();
          utils.submitFormCheckPlot({
            plotClass: 'hazard-lib-disagg-plot',
          });
          cy.get('.service-button').click();
          cy.get('nshmp-lib-url-bottom-sheet').contains('SA0P02=5');
        });
      });
    });
  });
});

function setControlPanel() {
  utils.setHazardCommonControlPanel();

  cy.get('nshmp-template-control-panel')
    .find('.imt-select')
    .find('mat-select')
    .click();
  utils.selectMatOption('PGA');
}
