import {Component, computed, OnDestroy, OnInit, Signal} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatFormFieldModule} from '@angular/material/form-field';
import {MatSelectModule} from '@angular/material/select';
import {
  NshmpLibNgHazardLocationFormComponent,
  NshmpLibNgHazardModelFormComponent,
  NshmpLibNgHazardSiteClassFormComponent,
} from '@ghsc/nshmp-lib-ng/hazard';
import {
  NshmpLibNgMapSelectSiteComponent,
  SelectSiteDialogData,
} from '@ghsc/nshmp-lib-ng/map';
import {
  NshmpLibNgControlPanelButtonsComponent,
  nshmpUtils,
  RETURN_PERIOD_BOUNDS,
  RETURN_PERIODS,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {Location} from '@ghsc/nshmp-utils-ts/libs/nshmp-lib/geo';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {Subscription} from 'rxjs';

import {AppService} from '../../services/app.service';
import {BuildingCodeControlComponent} from '../building-code-control/building-code-control.component';

@Component({
  imports: [
    NshmpLibNgHazardModelFormComponent,
    NshmpLibNgHazardLocationFormComponent,
    NshmpLibNgMapSelectSiteComponent,
    NshmpLibNgHazardSiteClassFormComponent,
    NshmpLibNgControlPanelButtonsComponent,
    MatFormFieldModule,
    ReactiveFormsModule,
    MatSelectModule,
    BuildingCodeControlComponent,
  ],
  selector: 'app-control-panel-hazard',
  styleUrl: './control-panel-hazard.component.scss',
  templateUrl: './control-panel-hazard.component.html',
})
export class ControlPanelHazardComponent implements OnInit, OnDestroy {
  private subs: Subscription[] = [];

  /** Form field state */
  formGroup = this.service.hazardFormGroup;
  /** List of imt `Parameter`s */
  imts = computed(() => this.service.hazardUsage()?.response?.models.imts);
  /** Max and min bounds for return periods */
  returnPeriodBounds = RETURN_PERIOD_BOUNDS;
  /** List of return periods */
  returnPeriods = RETURN_PERIODS;
  /** Data for select site component */
  selectSiteData: Signal<SelectSiteDialogData> = computed(() => {
    const usage = this.service.hazardUsage();
    const nshmService = this.service.nshmService();

    if (usage === null || nshmService === undefined) {
      return;
    }

    const services = environment.webServices.nshmpHazWs.services.curveServices;

    return {
      mapUrl: `${nshmService.url}${services.map}?raw=true`,
      sitesUrl: `${nshmService.url}${services.sites}?raw=true`,
    };
  });
  /** Site class select menu holder */
  siteClassPlaceHolder = nshmpUtils.selectPlaceHolder();
  /** List of site class `Parameter`s */
  siteClasses = computed(
    () => this.service.hazardUsage()?.response?.models.siteClasses,
  );

  constructor(public service: AppService) {}

  ngOnDestroy(): void {
    this.subs.forEach(sub => sub.unsubscribe());
  }

  ngOnInit(): void {
    const controls = this.formGroup.controls;

    this.subs.push(
      controls.latitude.valueChanges.subscribe(() => this.service.resetState()),
    );

    this.subs.push(
      controls.longitude.valueChanges.subscribe(() =>
        this.service.resetState(),
      ),
    );

    this.subs.push(
      controls.model.valueChanges.subscribe(() => this.onModelChange()),
    );

    this.subs.push(
      controls.imt.valueChanges.subscribe(() => this.service.hazardToRtgm()),
    );

    this.subs.push(
      controls.siteClass.valueChanges.subscribe(() =>
        this.service.hazardToRtgm(),
      ),
    );
  }

  onSubmit() {
    this.service.callHazardService();
  }

  setLocation(location: Location): void {
    this.service.hazardFormGroup.patchValue({
      latitude: location.latitude,
      longitude: location.longitude,
    });
  }

  private onModelChange(): void {
    const {latitude, longitude} = this.formGroup.getRawValue();
    const latitudeBounds = this.service.hazardUsage().response.models.latitude;
    const longitudeBounds =
      this.service.hazardUsage().response.models.longitude;

    if (
      !(
        latitude >= latitudeBounds.min &&
        latitude <= latitudeBounds.max &&
        longitude >= longitudeBounds.min &&
        longitude <= longitudeBounds.max
      )
    ) {
      this.formGroup.patchValue({
        latitude: NaN,
        longitude: NaN,
      });
    }

    this.formGroup.controls.latitude.markAsPristine();
    this.formGroup.controls.longitude.markAsPristine();
    this.service.resetState();
  }
}
