import {GmmImtFormControls} from '@ghsc/nshmp-lib-ng/gmm';
import {ServiceCallInfo} from '@ghsc/nshmp-lib-ng/nshmp';
import {
  NshmpPlot,
  NshmpPlotSettings,
  PlotOptions,
  plotUtils,
} from '@ghsc/nshmp-lib-ng/plot';
import {
  GmmDistanceResponse,
  GmmDistanceUsage,
} from '@ghsc/nshmp-utils-ts/libs/nshmp-ws/gmm-services';
import {EnumParameterValues} from '@ghsc/nshmp-utils-ts/libs/nshmp-ws-utils/metadata';

/**
 * Plot ids.
 */
export enum Plots {
  FAULT = 'FAULT',
  GROUND_MOTION = 'GROUND_MOTIONS',
}

/**
 * GMM Hanging Wall Effects state.
 */
export interface AppState {
  plots: Map<string, NshmpPlot>;
  /** Service call info */
  serviceCallInfo: ServiceCallInfo;
  /** GMM service response */
  serviceResponses: GmmDistanceResponse[];
  /** Supported IMTs */
  supportedImts: EnumParameterValues[];
  /** GMM usages */
  usageResponse: GmmDistanceUsage;
}

/**
 * Control panel form.
 */
export interface ControlPanelForm extends GmmImtFormControls {
  /** Max distance */
  rMax: number;
  /** Min distance */
  rMin: number;
}

/**
 * Returns the default plots.
 */
export function defaultPlots(): Map<string, NshmpPlot> {
  const plots = new Map<string, NshmpPlot>();

  plots.set(Plots.GROUND_MOTION, {
    label: 'Hanging Wall Effects',
    plotData: groundMotionPlotData,
    settingsForm: plotUtils.plotSettingsToFormGroup(groundMotionSettingsForm),
  });

  plots.set(Plots.FAULT, {
    label: 'Faults',
    plotData: faultPlotData,
    settingsForm: plotUtils.plotSettingsToFormGroup(faultSettingsForm),
  });

  return plots;
}

/** Plot options for ground motion plot */
const groundMotionPlotOptions: PlotOptions = {
  layout: {
    legend: {
      x: 0.99,
      xanchor: 'right',
      y: 0.98,
    },
    xaxis: {
      autorange: false,
      range: [-21, 101],
      type: 'linear',
    },
    yaxis: {
      range: [0, 1.0],
      type: 'linear',
    },
  },
};

/** Plot options for fault plot */
const faultPlotOptions: PlotOptions = {
  config: {
    displayModeBar: false,
  },
  layout: {
    aspectRatio: '2:1',
    showlegend: false,
    xaxis: {
      autorange: false,
      nticks: 20,
      range: [-21, 57.5],
      side: 'top',
      type: 'linear',
    },
    yaxis: {
      autorange: false,
      range: [40, 0],
      type: 'linear',
    },
  },
};

const groundMotionPlotData = plotUtils.defaultPlot({
  id: 'ground-motions',
  mobileOptions: {
    ...groundMotionPlotOptions,
    layout: {
      ...groundMotionPlotOptions.layout,
      aspectRatio: '4:3',
    },
  },
  options: groundMotionPlotOptions,
  title: 'Hanging Wall Effects',
  xLabel: 'Distance (km)',
  yLabel: 'Median ground motion (g)',
});

const groundMotionSettingsForm: NshmpPlotSettings = {
  config: groundMotionPlotData.config,
  layout: plotUtils.plotlyLayoutToSettings(groundMotionPlotData.layout),
};

const faultPlotData = plotUtils.defaultPlot({
  id: 'fault-id',
  mobileOptions: {
    ...faultPlotOptions,
    layout: {
      ...faultPlotOptions.layout,
      aspectRatio: '4:3',
    },
  },
  options: faultPlotOptions,
  panelBreakpoint: 660 * 0.6,
  title: '',
  xLabel: 'Distance (km)',
  yLabel: 'Depth (km)',
});

const faultSettingsForm: NshmpPlotSettings = {
  config: faultPlotData.config,
  layout: plotUtils.plotlyLayoutToSettings(faultPlotData.layout),
};
