import {AsyncPipe} from '@angular/common';
import {Component, computed, OnDestroy, OnInit} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatOption} from '@angular/material/core';
import {MatError, MatFormField, MatLabel} from '@angular/material/form-field';
import {MatInput} from '@angular/material/input';
import {MatSelect} from '@angular/material/select';
import {
  GmmFormControlIds,
  gmmUtils,
  MultiSelectableParam,
  MW_COMMON_VALUES,
  NshmpLibNgGmmMenuComponent,
  NshmpLibNgGmmMultiParamMenuComponent,
  NshmpLibNgGmmMultiSelectComponent,
  NshmpLibNgGmmPlotOptionsControlPanelComponent,
  VS30_COMMON_VALUES,
} from '@ghsc/nshmp-lib-ng/gmm';
import {
  NshmpLibNgControlPanelButtonsComponent,
  NshmpService,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {combineLatest, Subscription} from 'rxjs';
import {map} from 'rxjs/operators';

import {AppService} from '../../services/app.service';

/**
 * Control panel form fields and buttons to call GMM vs distance service.
 */
@Component({
  imports: [
    MatFormField,
    MatLabel,
    MatSelect,
    MatOption,
    MatInput,
    MatError,
    AsyncPipe,
    NshmpLibNgGmmMenuComponent,
    NshmpLibNgGmmMultiSelectComponent,
    NshmpLibNgGmmPlotOptionsControlPanelComponent,
    NshmpLibNgControlPanelButtonsComponent,
    NshmpLibNgGmmMultiParamMenuComponent,
    ReactiveFormsModule,
  ],
  selector: 'app-control-panel',
  styleUrl: './control-panel.component.scss',
  templateUrl: './control-panel.component.html',
})
export class ControlPanelComponent implements OnInit, OnDestroy {
  /** List of common magnitudes */
  MwCommonValues = MW_COMMON_VALUES;
  /** List of common vs30 values */
  vs30CommonValues = VS30_COMMON_VALUES;

  /** Form controls state */
  controls = this.service.formGroup.controls;

  /** Control panel form field state */
  form = this.service.formGroup;

  /** Whether a GMM has been selected */
  gmmSelected = this.controls.gmmSource.getRawValue()?.length > 0;

  /** Usage parameters */
  parameters = computed(() => this.service.usage()?.response.parameters);

  /** Service call info state */
  serviceCallInfo = this.service.serviceCallInfo;

  /** Supported IMTs based on GMMs selected */
  supportedImts = this.service.supportedImts;

  /** Whether magnitudes are multi-selectable */
  MwMultiple$ = this.controls.multiSelectableParam.valueChanges.pipe(
    map(
      multiSelectableParam =>
        multiSelectableParam.toString() === GmmFormControlIds.MW.toString(),
    ),
  );

  /** Whether Vs30 is multi-selectable */
  vs30Multiple$ = this.controls.multiSelectableParam.valueChanges.pipe(
    map(
      multiSelectableParam =>
        multiSelectableParam.toString() === GmmFormControlIds.VS30.toString(),
    ),
  );

  private subs: Subscription[] = [];

  constructor(
    public service: AppService,
    private nshmpService: NshmpService,
  ) {}

  ngOnInit(): void {
    this.subs.push(
      this.controls.gmmSource.valueChanges.subscribe(() => {
        this.onGmmSource();
      }),
    );

    this.subs.push(
      this.controls.multiSelectableParam.valueChanges.subscribe(() =>
        this.onMultiSelectableParam(),
      ),
    );

    this.subs.push(
      this.controls.showEpistemicUncertainty.valueChanges.subscribe(() =>
        this.onShowEpistemicUncertainty(),
      ),
    );

    this.subs.push(
      combineLatest([
        this.controls.dip.valueChanges,
        this.controls.imt.valueChanges,
        this.controls.Mw.valueChanges,
        this.controls.MwMulti.valueChanges,
        this.controls.vs30Multi.valueChanges,
        this.controls.vs30.valueChanges,
        this.controls.z1p0.valueChanges,
        this.controls.z2p5.valueChanges,
        this.controls.zSed.valueChanges,
        this.controls.width.valueChanges,
      ]).subscribe(() => this.service.resetState()),
    );
  }

  ngOnDestroy(): void {
    this.subs.forEach(sub => sub?.unsubscribe());
  }

  /**
   * On form submit.
   */
  onSubmit(): void {
    this.service.callService();
    this.nshmpService.selectPlotControl();
  }

  private onGmmSource() {
    const formValues = this.form.getRawValue();

    if (
      !this.service.state().usageResponse ||
      formValues.gmmSource.length === 0
    ) {
      this.controls.imt.patchValue(gmmUtils.imtPlaceHolder().value);
      this.service.updateState({
        supportedImts: [],
      });
      return;
    }

    const supportedImts = gmmUtils.getSupportedImts(
      formValues.gmmSource,
      this.service.state().usageResponse,
    );

    this.service.updateState({
      supportedImts,
    });

    const imt = [...supportedImts].shift();

    this.controls.imt.patchValue(imt?.value);
    this.service.resetState();
  }

  private onMultiSelectableParam(): void {
    if (!this.service.state().usageResponse) {
      return;
    }

    const {multiSelectableParam} = this.form.getRawValue();
    const parameters = this.service.state().usageResponse.response.parameters;

    if (multiSelectableParam === MultiSelectableParam.MW) {
      this.controls.Mw.setValue(parameters.Mw.value as number);
    } else if (multiSelectableParam === MultiSelectableParam.VS30) {
      this.controls.vs30.setValue(parameters.vs30.value as number);
    }

    this.controls.gmmSource.setValue([]);
    this.controls.vs30Multi.setValue([]);
    this.controls.MwMulti.setValue([]);
    this.service.resetState();
  }

  private onShowEpistemicUncertainty(): void {
    this.service.createPlots();
  }
}
