import {Component, computed, OnDestroy, OnInit} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {
  MatButtonToggle,
  MatButtonToggleGroup,
} from '@angular/material/button-toggle';
import {MatCheckbox} from '@angular/material/checkbox';
import {MatError, MatFormField, MatLabel} from '@angular/material/form-field';
import {MatInput} from '@angular/material/input';
import {Subscription} from 'rxjs';

import {AppService} from '../../services/app.service';

/**
 * Control panel form fields for GMM path parameters.
 */
@Component({
  imports: [
    MatLabel,
    MatFormField,
    MatInput,
    MatError,
    MatCheckbox,
    MatButtonToggleGroup,
    MatButtonToggle,
    ReactiveFormsModule,
  ],
  selector: 'app-path-parameters',
  styleUrl: './path-parameters.component.scss',
  templateUrl: './path-parameters.component.html',
})
export class PathParametersComponent implements OnInit, OnDestroy {
  /** Control panel form field state */
  form = this.service.formGroup;

  /** Usage parameters */
  parameters = computed(() => this.service.usage()?.response.parameters);

  subs: Subscription[] = [];

  private controls = this.form.controls;

  constructor(private service: AppService) {}

  ngOnInit(): void {
    this.subs.push(this.controls.rX.valueChanges.subscribe(() => this.onRx()));

    this.subs.push(
      this.controls.derivePathParams.valueChanges.subscribe(() =>
        this.onDerivePathParams(),
      ),
    );

    this.subs.push(
      this.controls.hangingWall.valueChanges.subscribe(() =>
        this.onHangingWall(),
      ),
    );
  }

  ngOnDestroy(): void {
    this.subs.forEach(sub => sub.unsubscribe());
  }

  private onDerivePathParams(): void {
    const {derivePathParams, hangingWall, rX} = this.form.getRawValue();

    if (derivePathParams) {
      if (rX >= 0 !== hangingWall) {
        this.controls.hangingWall.setValue(rX >= 0);
      }
      this.controls.hangingWall.enable();
      this.controls.rJB.disable();
      this.controls.rRup.disable();
      this.service.calcPathParameters(this.form);
    } else {
      this.controls.hangingWall.setValue(null);
      this.controls.hangingWall.disable();
      this.controls.rJB.enable();
      this.controls.rRup.enable();
    }

    this.service.resetState();
  }

  private onHangingWall(): void {
    const values = this.form.getRawValue();
    const rX = values.hangingWall ? Math.abs(values.rX) : -Math.abs(values.rX);

    if (values.rX !== rX) {
      this.controls.rX.setValue(rX);
    }
    this.service.calcPathParameters(this.form);
    this.service.resetState();
  }

  private onRx(): void {
    const {derivePathParams, hangingWall, rX} = this.form.getRawValue();

    if (derivePathParams && rX >= 0 !== hangingWall) {
      this.controls.hangingWall.patchValue(rX >= 0);
    }

    this.service.calcPathParameters(this.form);
    this.service.resetState();
  }
}
