import {AsyncPipe, DatePipe, NgClass, UpperCasePipe} from '@angular/common';
import {
  AfterViewInit,
  Component,
  effect,
  ElementRef,
  OnDestroy,
  viewChild,
} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatButtonToggleGroup} from '@angular/material/button-toggle';
import {MatIcon} from '@angular/material/icon';
import {MatProgressSpinner} from '@angular/material/progress-spinner';
import {
  MatSlideToggle,
  MatSlideToggleChange,
} from '@angular/material/slide-toggle';
import {MatSnackBar} from '@angular/material/snack-bar';
import {MatToolbar, MatToolbarRow} from '@angular/material/toolbar';
import {MatTooltip} from '@angular/material/tooltip';
import {JobStatus} from '@ghsc/nshmp-lib-ng/aws';
import {NshmpService} from '@ghsc/nshmp-lib-ng/nshmp';
import {NshmpTemplateService} from '@ghsc/nshmp-template';

import {GetLogsResponseData} from '../../models/get-logs.model';
import {LogPipe} from '../../pipes/log.pipe';
import {AppService} from '../../services/app.service';
import {SnackBarComponent} from '../snack-bar/snack-bar.component';

/**
 * Main application content showing logs.
 */
@Component({
  imports: [
    MatToolbar,
    MatToolbarRow,
    MatProgressSpinner,
    MatSlideToggle,
    MatTooltip,
    NgClass,
    UpperCasePipe,
    DatePipe,
    LogPipe,
    ReactiveFormsModule,
    MatIcon,
    MatButtonToggleGroup,
    AsyncPipe,
  ],
  selector: 'app-content',
  styleUrl: './content.component.scss',
  templateUrl: './content.component.html',
})
export class ContentComponent implements OnDestroy, AfterViewInit {
  /** Job status enum */
  jobStatus = JobStatus;

  /** Element reference to logs element */
  readonly logsEl = viewChild<ElementRef<HTMLDivElement>>('logs');

  /** The log service response */
  serviceResponse = this.service.serviceResponse;
  /** Control form state */
  form = this.service.formGroup;
  /** Whether service is loading */
  loading = this.service.serviceLoading;

  constructor(
    public templateServie: NshmpTemplateService,
    private service: AppService,
    private snackBar: MatSnackBar,
    private nshmpService: NshmpService,
  ) {
    effect(() => {
      const serviceResponse = this.service.serviceResponse();

      if (serviceResponse) {
        this.scrollBottom();
        this.checkLogs(serviceResponse.response);
      }
    });
  }

  ngOnDestroy(): void {
    this.snackBar.dismiss();
  }

  ngAfterViewInit(): void {
    this.scrollBottom();
  }

  /**
   * Handle slide toogle change for application auto refreshe
   */
  onRefreshChange(event: MatSlideToggleChange): void {
    if (event.checked) {
      this.service.callService(false);
    }
  }

  /**
   * Scroll to bottom of page for logs.
   */
  scrollBottom(): void {
    /* Wait for log to display and scroll */
    setTimeout(() => {
      const logsEl = this.logsEl();
      if (logsEl) {
        logsEl.nativeElement.scroll({
          top: logsEl.nativeElement.scrollHeight,
        });
      }
    }, 0);
  }

  /**
   * Scroll to top of logs.
   */
  scrollTop() {
    this.logsEl().nativeElement.scroll({top: 0});
  }

  /**
   * Check log response for errors or results.
   *
   * @param response The logs service response
   */
  private checkLogs(response: GetLogsResponseData): void {
    if (
      response.jobStatus === JobStatus.FAILED &&
      this.service.logHasForSupport(response)
    ) {
      if (response.results) {
        this.snackBar.openFromComponent(SnackBarComponent, {
          data: response.results,
        });
      }
      this.nshmpService.throwError$(
        new Error('nshmp-haz failed, check email for more information'),
      );
    } else if (response.results && this.service.logHasFinished(response)) {
      this.snackBar.openFromComponent(SnackBarComponent, {
        data: response.results,
      });
    }
  }
}
