import {Location as LocationService} from '@angular/common';
import {HttpParams} from '@angular/common/http';
import {computed, Injectable, Signal, signal} from '@angular/core';
import {FormBuilder, Validators} from '@angular/forms';
import {ActivatedRoute} from '@angular/router';
import {
  FormGroupControls,
  NshmpService,
  ServiceCallInfo,
  SpinnerService,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {
  DefaultPlotOptions,
  NshmpPlot,
  PlotOptions,
  plotUtils,
} from '@ghsc/nshmp-lib-ng/plot';
import {
  GeophysicalResponse,
  GeophysicalUsage,
} from '@ghsc/nshmp-utils-ts/libs/ncm/geophysical';
import {Location} from '@ghsc/nshmp-utils-ts/libs/nshmp-lib/geo';
import deepEqual from 'deep-equal';
import {PlotData} from 'plotly.js';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {AppServiceModel} from 'projects/nshmp-apps/src/shared/models/app-service.model';
import {SharedService} from 'projects/nshmp-apps/src/shared/services/shared.service';
import {apps} from 'projects/nshmp-apps/src/shared/utils/applications.utils';
import {catchError} from 'rxjs';

import {ControlForm} from '../models/control-form.model';
import {Plots} from '../models/plots.model';
import {AppState} from '../models/state.model';

/**
 * URL query parameters
 */
interface Query {
  depthInc?: number;
  depthMax?: number;
  depthMin?: number;
  location?: string | string[];
}

/**
 * NCM geophysical profiles application service.
 */
@Injectable({
  providedIn: 'root',
})
export class AppService
  extends SharedService
  implements AppServiceModel<AppState, ControlForm>
{
  private ncm = environment.webServices.ncm;
  private serviceUrl = `${this.ncm.url}${this.ncm.services.geophysical}`;

  /** Form group */
  formGroup = this.formBuilder.group({
    ...this.defaultFormValues(),
    locations: this.formBuilder.array<FormGroupControls<Location>>([]),
  });

  /** State */
  state = signal<AppState>(this.initialState());

  constructor(
    private formBuilder: FormBuilder,
    private spinnerService: SpinnerService,
    private nshmpService: NshmpService,
    private route: ActivatedRoute,
    private location: LocationService,
  ) {
    super();

    this.addValidators();
  }

  /**
   * Returns the plots.
   */
  get plots(): Signal<Plots> {
    return computed(() => this.state().plots);
  }

  /**
   * Returns the service call info.
   */
  get serviceCallInfo(): Signal<ServiceCallInfo> {
    return computed(() => this.state().serviceCallInfo);
  }

  /**
   * Returns the service response.
   */
  get serviceResponse(): Signal<GeophysicalResponse> {
    return computed(() => this.state().serviceResponse);
  }

  /**
   * Returns the usage response.
   */
  get usage(): Signal<GeophysicalUsage> {
    return computed(() => this.state().usageResponse);
  }

  /**
   * Add a new location form group to form array.
   */
  addLocationForm(): void {
    this.formGroup.controls.locations.controls.push(this.locationForm());
    this.formGroup.controls.locations.updateValueAndValidity();
  }

  /**
   * Add form validators
   */
  addValidators(): void {
    const controls = this.formGroup.controls;

    controls.depthInc.addValidators(this.validateRequired());
    controls.depthMax.addValidators(this.validateRequired());
    controls.depthMin.addValidators(this.validateRequired());
    controls.locations.addValidators(this.validateRequired());
  }

  /**
   * Call NCM geophysical web service
   */
  callService(): void {
    const spinnerRef = this.spinnerService.show(SpinnerService.MESSAGE_SERVICE);

    const values = this.formGroup.getRawValue();
    const params = new HttpParams().appendAll({
      depths: `${values.depthMin},${values.depthInc},${values.depthMax}`,
      location: this.formGroup.controls.locations
        .getRawValue()
        .map(location => `${location.latitude},${location.longitude}`)
        .join(','),
    });
    const url = `${this.serviceUrl}?${params.toString()}`;

    this.nshmpService
      .callService$<GeophysicalResponse>(url)
      .pipe(
        catchError((error: Error) => {
          spinnerRef.close();
          return this.nshmpService.throwError$(error);
        }),
      )
      .subscribe(serviceResponse => {
        this.updateState({
          serviceCallInfo: {
            ...this.state().serviceCallInfo,
            serviceCalls: [url],
          },
          serviceResponse,
        });

        this.createPlots();
        spinnerRef.close();
      });
  }

  /**
   * Returns the default control form values.
   */
  defaultFormValues(): ControlForm {
    return {
      depthInc: 100,
      depthMax: 2000,
      depthMin: 0,
      locations: [],
    };
  }

  /**
   * Returns the default plots.
   */
  defaultPlots(): Plots {
    const yLabel = 'Depth (meters)';

    return {
      density: this.defaultPlot({
        id: 'density',
        title: 'Density Profile',
        xLabel: 'Density Profile (kg/m3)',
        yLabel,
      }),
      porosity: this.defaultPlot({
        id: 'porosity',
        title: 'Porosity Profile',
        xLabel: 'Porosity Profile',
        yLabel,
      }),
      temperature: this.defaultPlot({
        id: 'temperature',
        title: 'Temperature Profile',
        xLabel: 'Temperature Profile (°C)',
        yLabel,
      }),
      vp: this.defaultPlot({
        id: 'p-wave',
        title: 'P-wave Profile',
        xLabel: 'P-wave Profile (m/s)',
        yLabel,
      }),
      vs: this.defaultPlot({
        id: 's-wave',
        title: 'S-wave Profile',
        xLabel: 'S-wave Profile (m/s)',
        yLabel,
      }),
    };
  }

  /**
   * Initialize application, call usage.
   */
  init(): void {
    this.formGroup.controls.locations.clear();

    const spinnerRef = this.spinnerService.show(
      SpinnerService.MESSAGE_METADATA,
    );

    this.nshmpService
      .callService$<GeophysicalUsage>(this.serviceUrl)
      .pipe(
        catchError((error: Error) => {
          spinnerRef.close();
          return this.nshmpService.throwError$(error);
        }),
      )
      .subscribe(usageResponse => {
        spinnerRef.close();

        this.updateState({
          serviceCallInfo: {
            ...this.state().serviceCallInfo,
            usage: [this.serviceUrl],
          },
          usageResponse,
        });

        this.initialFormSet();
      });
  }

  /**
   * Returns the initial state.
   */
  initialState(): AppState {
    return {
      plots: this.defaultPlots(),
      serviceCallInfo: {
        serviceCalls: [],
        serviceName: 'Geophysical profiles',
        usage: [],
      },
      serviceResponse: null,
      usageResponse: null,
    };
  }

  /**
   * Reset the control panel.
   */
  resetControlPanel(): void {
    this.formGroup.reset({
      ...this.defaultFormValues(),
    });

    this.formGroup.controls.locations.clear();
    this.addLocationForm();

    this.resetState();
  }

  /**
   * Reset the current state.
   */
  resetState(): void {
    this.updateState({
      plots: this.defaultPlots(),
      serviceCallInfo: {
        ...this.state().serviceCallInfo,
        serviceCalls: [],
      },
      serviceResponse: null,
    });
  }

  /**
   * Update the state.
   *
   * @param state The partial state to update.
   */
  updateState(state: Partial<AppState>): void {
    const updatedState = {
      ...this.state(),
      ...state,
    };

    if (!deepEqual(updatedState, this.state())) {
      this.state.set({
        ...this.state(),
        ...state,
      });
    }
  }

  /**
   * Create the plots for service response.
   */
  private createPlots(): void {
    this.updateState({
      plots: this.serviceResponseToPlots(),
    });
  }

  /**
   * Returns a default plot.
   *
   * @param options The plot options
   */
  private defaultPlot(options: DefaultPlotOptions): NshmpPlot {
    const values = this.formGroup.getRawValue();

    const plotOptions: PlotOptions = {
      layout: {
        xaxis: {
          nticks: 15,
          range: [0, 100],
          type: 'linear',
        },
        yaxis: {
          autorange: 'reversed',
          nticks: 15,
          range: [values.depthMin, values.depthMax],
          type: 'linear',
        },
      },
    };

    const plotData = plotUtils.defaultPlot({
      ...options,
      mobileOptions: plotOptions,
      options: plotOptions,
    });

    return {
      label: options.title,
      plotData,
      settingsForm: plotUtils.plotSettingsToFormGroup({
        config: plotData.config,
        layout: plotUtils.plotlyLayoutToSettings(plotData.layout),
      }),
    };
  }

  /**
   * Initial form set from query parameters.
   */
  private initialFormSet(): void {
    const query = this.route.snapshot.queryParams as Query;
    const locations: Location[] = [];

    if (query.location) {
      if (Array.isArray(query.location)) {
        locations.push(...query.location.map(loc => this.queryToLocation(loc)));
      } else {
        locations.push(this.queryToLocation(query.location));
      }
    } else {
      this.addLocationForm();
    }

    locations.forEach(() => {
      this.addLocationForm();
    });

    const defaultValues = this.defaultFormValues();

    this.formGroup.patchValue({
      depthInc: query.depthInc ?? defaultValues.depthInc,
      depthMax: query.depthMax ?? defaultValues.depthMax,
      depthMin: query.depthMin ?? defaultValues.depthMin,
      locations,
    });

    if (this.formGroup.valid) {
      this.callService();
    } else {
      this.formGroup.markAsDirty();
    }
    this.formGroup.valueChanges.subscribe(() => this.updateUrl());
  }

  /**
   * Returns a new location form group for the form array.
   */
  private locationForm(): FormGroupControls<Location> {
    const form = this.formBuilder.group({
      latitude: [NaN, [this.validateRequired(), this.validateNan()]],
      longitude: [NaN, [this.validateRequired(), this.validateNan()]],
    });

    if (this.usage()) {
      const locations = this.usage().response.parameters.location;

      form.controls.latitude.addValidators([
        Validators.min(locations.latitude.min),
        Validators.max(locations.latitude.max),
      ]);
      form.controls.longitude.addValidators([
        Validators.min(locations.longitude.min),
        Validators.max(locations.longitude.max),
      ]);
    }

    form.valueChanges.subscribe(() => {
      this.formGroup.controls.locations.updateValueAndValidity();
      this.formGroup.controls.locations.markAsDirty();
    });

    return form;
  }

  /**
   * Convert string comma delimited latitude and longitude into Location.
   *
   * @param location The comma delimited location: "latitude, longitude"
   * @returns
   */
  private queryToLocation(location: string): Location {
    const [latitude, longitude] = location.split(',');

    return {
      latitude: Number.parseFloat(latitude),
      longitude: Number.parseFloat(longitude),
    };
  }

  /**
   * Transform the service response to plots.
   */
  private serviceResponseToPlots(): Plots {
    const serviceResponse = this.state().serviceResponse;

    if (serviceResponse === null || serviceResponse === undefined) {
      return this.defaultPlots();
    }

    const temperaturePlotData: Partial<PlotData>[] = [];
    const porosityPlotData: Partial<PlotData>[] = [];
    const vsPlotData: Partial<PlotData>[] = [];
    const vpPlotData: Partial<PlotData>[] = [];
    const densityPlotData: Partial<PlotData>[] = [];

    const ys = serviceResponse.request.depths.depth_vector;

    serviceResponse.response.results.forEach(result => {
      const name = `${result.location.latitude}, ${result.location.longitude}`;

      temperaturePlotData.push({
        mode: 'lines+markers',
        name,
        x: result.profile.temperature,
        y: ys,
      });

      porosityPlotData.push({
        mode: 'lines+markers',
        name,
        x: result.profile.porosity,
        y: ys,
      });

      vpPlotData.push({
        mode: 'lines+markers',
        name,
        x: result.profile.vp,
        y: ys,
      });

      vsPlotData.push({
        mode: 'lines+markers',
        name,
        x: result.profile.vs,
        y: ys,
      });

      densityPlotData.push({
        mode: 'lines+markers',
        name,
        x: result.profile.density,
        y: ys,
      });
    });

    const plots = this.plots();

    return {
      density: {
        ...plots.density,
        plotData: {
          ...plots.density.plotData,
          data: densityPlotData,
        },
      },
      porosity: {
        ...plots.porosity,
        plotData: {
          ...plots.porosity.plotData,
          data: porosityPlotData,
        },
      },
      temperature: {
        ...plots.temperature,
        plotData: {
          ...plots.temperature.plotData,
          data: temperaturePlotData,
        },
      },
      vp: {
        ...plots.vp,
        plotData: {
          ...plots.vp.plotData,
          data: vpPlotData,
        },
      },
      vs: {
        ...plots.vs,
        plotData: {
          ...plots.vs.plotData,
          data: vsPlotData,
        },
      },
    };
  }

  /**
   * Update URL query based on control panel change.
   */
  private updateUrl(): void {
    const values = this.formGroup.getRawValue();

    const location = this.formGroup.controls.locations.controls
      .map(control => control.getRawValue())
      .map(value => `${value.latitude},${value.longitude}`);

    this.location.replaceState(
      apps().ncm.geophysicalProfiles.routerLink,
      new HttpParams()
        .appendAll({
          depthInc: values.depthInc,
          depthMax: values.depthMax,
          depthMin: values.depthMin,
          location,
        })
        .toString(),
    );
  }
}
